"""Tests for :class:`.BaseConsumer`."""

from unittest import TestCase, mock
import time
from botocore.exceptions import BotoCoreError, WaiterError, ClientError

from ..consumer import BaseConsumer, CheckpointError, StopProcessing, \
    KinesisRequestFailed, ConfigurationError, StreamNotAvailable, \
    process_stream, RestartProcessing


class TestBaseConsumer(TestCase):
    """Test :class:`.BaseConsumer` behavior and public methods."""

    def setUp(self):
        """Create a mock checkpointer."""
        self.checkpointer = mock.MagicMock()
        self.checkpointer.position = None

    @mock.patch('boto3.Session')
    def test_go(self, mock_session_factory):
        """On go, consumer should wait for stream to be available."""
        mock_client = mock.MagicMock()
        mock_waiter = mock.MagicMock()
        mock_client.get_waiter.return_value = mock_waiter
        mock_client.get_records.side_effect = StopProcessing
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))

        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer, delay=0)
        consumer.sleep_time = 0     # Don't wait.
        try:
            consumer.go()
        except StopProcessing:
            pass    # This was intentional; we needed to stop processing.
        except Exception:
            self.fail('If the waiter returns without an exception, no'
                      ' exception should be raised.')
        self.assertEqual(mock_waiter.wait.call_count, 1,
                         "A boto3 waiter should be used")

    @mock.patch('boto3.Session')
    def test_go_stream_not_available(self, mock_session_factory):
        """If the stream is not available, should raise an exception."""
        mock_client = mock.MagicMock()
        mock_waiter = mock.MagicMock()

        def raise_waiter_error(*a, **k):
            raise WaiterError('', {}, {})

        mock_waiter.wait.side_effect = raise_waiter_error
        mock_client.get_waiter.return_value = mock_waiter
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer, delay=0)
        consumer.sleep_time = 0     # Don't wait.
        with self.assertRaises(StreamNotAvailable):
            consumer.go()

    @mock.patch('boto3.Session')
    def test_iteration(self, mock_session_factory):
        """Test iteration behavior."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_records.return_value = {
            'Records': [
                {'SequenceNumber': str(i)} for i in range(10)
            ],
            'NextShardIterator': '10'
        }
        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer)
        consumer.sleep_time = 0     # Don't wait.
        consumer.client = consumer.new_client()

        next_start, processed = consumer.process_records('0')
        self.assertGreater(mock_client.get_records.call_count, 0)
        self.assertEqual(processed, 10)
        self.assertEqual(next_start, '10', "Should return NextShardIterator")

    @mock.patch('boto3.Session')
    def test_process_records_until_shard_closes(self, mock_session_factory):
        """Should call GetRecords until no next iterator is available."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}

        def get_records(**kwargs):
            start = int(kwargs['ShardIterator'])
            end = start + int(kwargs['Limit'])
            if start > 100:
                return {'Records': [], 'NextShardIterator': None}
            return {
                'Records': [
                    {'SequenceNumber': str(i)} for i in range(start, end)
                ],
                'NextShardIterator': str(end + 1)
            }

        mock_client.get_records.side_effect = get_records

        batch_size = 50
        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer,
                                batch_size=batch_size)
        consumer.sleep_time = 0     # Don't wait.
        with self.assertRaises(StopProcessing):
            consumer.go()
        self.assertEqual(mock_client.get_records.call_count,
                         (100/batch_size) + 1,
                         "Should call Kinesis GetRecords until no iterator"
                         " is returned.")

    @mock.patch('boto3.Session')
    def test_process_records_with_clienterror(self, mock_session_factory):
        """Should try to checkpoint before exiting."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}

        def raise_client_error(*args, **kwargs):
            raise ClientError({'Error': {'Code': 'foo'}}, {})

        mock_client.get_records.side_effect = raise_client_error

        batch_size = 50
        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer,
                                batch_size=batch_size, delay=0)
        consumer.sleep_time = 0     # Don't wait.
        consumer.position = 'fooposition'
        try:
            consumer.go()
        except Exception:
            pass
        self.assertEqual(self.checkpointer.checkpoint.call_count, 1)

    @mock.patch('boto3.Session')
    def test_start_from_timestamp(self, mock_session_factory):
        """Consumer is initialized with start_type 'AT_TIMESTAMP'."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}

        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer,
                                start_type='AT_TIMESTAMP')
        consumer.sleep_time = 0     # Don't wait.
        consumer.client = consumer.new_client()
        consumer._get_iterator()
        args, kwargs = mock_client.get_shard_iterator.call_args
        self.assertEqual(kwargs['ShardIteratorType'], 'AT_TIMESTAMP')
        self.assertIn('Timestamp', kwargs)

    @mock.patch('boto3.Session')
    def test_start_from_position(self, mock_session_factory):
        """Consumer is initialized with start_type 'AT_TIMESTAMP'."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}

        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer,
                                start_type='AT_TIMESTAMP')
        consumer.sleep_time = 0     # Don't wait.
        consumer.client = consumer.new_client()
        consumer.position = 'fooposition'
        consumer._get_iterator()
        args, kwargs = mock_client.get_shard_iterator.call_args
        self.assertEqual(kwargs['ShardIteratorType'], 'AFTER_SEQUENCE_NUMBER')
        self.assertEqual(kwargs['StartingSequenceNumber'], 'fooposition')

    @mock.patch('boto3.Session')
    def test_start_from_trim_horizon(self, mock_session_factory):
        """Consumer is initialized with start_type 'AT_TIMESTAMP'."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}

        consumer = BaseConsumer('foo', '1', 'a1b2c3d4', 'qwertyuiop',
                                'us-east-1', self.checkpointer,
                                start_type='TRIM_HORIZON')
        consumer.sleep_time = 0     # Don't wait.
        consumer.client = consumer.new_client()
        consumer._get_iterator()
        args, kwargs = mock_client.get_shard_iterator.call_args
        self.assertEqual(kwargs['ShardIteratorType'], 'TRIM_HORIZON')
        self.assertNotIn('StartingSequenceNumber', kwargs)


class TestProcessStream(TestCase):
    """Tests for :func:`.process_stream`."""

    def setUp(self):
        """Define a testing config."""
        self.config = {
            'KINESIS_STREAM': 'fooStream',
            'KINESIS_SHARD_ID': 'shard-0000000',
            'AWS_ACCESS_KEY_ID': 'ack',
            'AWS_SECRET_ACCESS_KEY': 'qwerty',
            'AWS_REGION': 'su-tsae-9'
        }

    @mock.patch('boto3.Session')
    def test_process_stream(self, mock_session_factory):
        """Run :func:`.process_stream` with a vanilla config."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}
        mock_client.get_records.return_value = {
            "Records": [
                {'SequenceNumber': '1', 'Data': 'bat'},
                {'SequenceNumber': '2', 'Data': 'abt'},
                {'SequenceNumber': '3', 'Data': 'tab'},
                {'SequenceNumber': '4', 'Data': 'bta'},
            ],
            "NextShardIterator": "5"
        }

        class FooConsumer(BaseConsumer):
            def __init__(self, *args, **kwargs):
                super(FooConsumer, self).__init__(*args, **kwargs)
                self.sleep_time = 0.01

            process_record = mock.MagicMock()

        class FooCheckpointer(object):
            def __init__(self):
                self.position = None

            def checkpoint(self, position):
                self.position = position
        start = time.time()
        process_stream(FooConsumer, self.config, FooCheckpointer(), 2)
        self.assertGreaterEqual(time.time() - start, 2,
                                "Should run for at least 2 seconds")
        self.assertGreater(FooConsumer.process_record.call_count, 0,
                           "Should be called at least several times")

    @mock.patch('boto3.Session')
    def test_restart_processing(self, mock_session_factory):
        """The record processor raises :class:`.RestartProcessing`."""
        mock_client = mock.MagicMock()
        mock_session_factory.return_value = \
            mock.MagicMock(client=mock.MagicMock(return_value=mock_client))
        mock_client.get_shard_iterator.return_value = {'ShardIterator': '1'}
        mock_client.get_records.return_value = {
            "Records": [
                {'SequenceNumber': '1', 'Data': 'bat'},
                {'SequenceNumber': '2', 'Data': 'abt'},
                {'SequenceNumber': '3', 'Data': 'tab'},
                {'SequenceNumber': '4', 'Data': 'bta'},
            ],
            "NextShardIterator": "5"
        }

        class FooConsumer(BaseConsumer):
            def __init__(self, *args, **kwargs):
                super(FooConsumer, self).__init__(*args, **kwargs)
                self.sleep_time = 0.01

            go = mock.MagicMock(side_effect=[RestartProcessing,
                                             RestartProcessing,
                                             StopProcessing])

        class FooCheckpointer(object):
            def __init__(self):
                self.position = None

            def checkpoint(self, position):
                self.position = position

        process_stream(FooConsumer, self.config, FooCheckpointer(), 2)
        self.assertEqual(FooConsumer.go.call_count, 3,
                         'Should be retried')



"""Tests for :mod:`.middleware`."""

from unittest import TestCase, mock
from flask import Flask

import sys


class TestWrap(TestCase):
    """Test :func:`.middleware.wrap`."""

    def setUp(self):
        """Define a minimal request environment."""
        self.environ = {
            'SERVER_NAME': 'fooserver',
            'SERVER_PORT': '123',
            'wsgi.url_scheme': 'http',
            'REQUEST_METHOD': 'HEAD'
        }

    def test_wrap_attaches_middleware(self):
        """:func:`.middleware.wrap` attaches WSGI middleware to a Flask app."""
        from arxiv.base import middleware

        class FooMiddleware(object):
            def __init__(self, app):
                self.app = app
                self.called = False

            def __call__(self, environ, start_response):
                self.called = True
                return self.app(environ, start_response)

            @property
            def wsgi_app(self):
                return self

        app = Flask('test')
        middleware.wrap(app, [FooMiddleware])
        self.assertIsInstance(app.wsgi_app, FooMiddleware,
                              "Middleware should be attached to Flask app")
        app(self.environ, mock.MagicMock())
        self.assertTrue(app.wsgi_app.called,
                        "Middleware should be called when Flask app is called")

    def test_wrap_is_inside_out(self):
        """Order of middleware determines call order upon request."""
        from arxiv.base import middleware

        class FirstMiddleware(middleware.base.BaseMiddleware):
            def before(self, environ, start_response):
                environ['call_order'].append('first')
                return environ, start_response

        class SecondMiddleware(middleware.base.BaseMiddleware):
            def before(self, environ, start_response):
                environ['call_order'].append('second')
                return environ, start_response

        app = Flask('test')
        self.environ['call_order'] = []
        middleware.wrap(app, [FirstMiddleware, SecondMiddleware])
        app(self.environ, mock.MagicMock())
        self.assertEqual(self.environ['call_order'][0], 'first',
                         "The first middleware should be called first")
        self.assertEqual(self.environ['call_order'][1], 'second',
                         "The second middleware should be called second")


class TestBaseMiddleware(TestCase):
    """Tests :class:`.middleware.base.BaseMiddleware`."""

    def setUp(self):
        """Define a minimal request environment."""
        self.environ = {
            'SERVER_NAME': 'fooserver',
            'SERVER_PORT': '123',
            'wsgi.url_scheme': 'http',
            'REQUEST_METHOD': 'HEAD'
        }

    def test_base(self):
        """:class:`.BaseMiddleware` doesn't do much."""
        from arxiv.base.middleware.base import BaseMiddleware
        app = Flask('test')
        bm = BaseMiddleware(app.wsgi_app)
        app.wsgi_app = bm
        app(self.environ, mock.MagicMock())
        self.assertEqual(bm, bm.wsgi_app)


class TestRequestLogs(TestCase):
    """Tests :class:`.middleware.request_logs.ClassicLogsMiddleware`."""

    def setUp(self):
        """Define a minimal request environment."""
        self.environ = {
            'SERVER_NAME': 'fooserver',
            'SERVER_PORT': '123',
            'wsgi.url_scheme': 'http',
            'REQUEST_METHOD': 'HEAD'
        }

    def test_uwsgi_is_available(self):
        """Attach logging middleware to app when uwsgi is available."""
        mock_uwsgi = mock.MagicMock()
        sys.modules['uwsgi'] = mock_uwsgi
        from arxiv.base.middleware import request_logs, wrap

        app = Flask('test')
        wrap(app, [request_logs.ClassicLogsMiddleware])
        self.assertIsInstance(app.wsgi_app, request_logs.ClassicLogsMiddleware,
                              "ClassicLogsMiddleware should be the outermost"
                              " middleware")
        app(self.environ, mock.MagicMock())
        self.assertGreater(mock_uwsgi.set_logvar.call_count, 0,
                           "Should set logging variables for uwsgi")

    def test_uwsgi_is_not_available(self):
        """Attach base middleware when uwsgi is not available."""
        if 'uwsgi' in sys.modules:
            del sys.modules['uwsgi']
        from arxiv.base.middleware import request_logs, wrap
        app = Flask('test')
        wrap(app, [request_logs.ClassicLogsMiddleware])
        self.assertIsInstance(app.wsgi_app, request_logs.BaseMiddleware,
                              "BaseMiddleware is attached instead")



"""Tests for :mod:`arxiv.forms`."""

from unittest import TestCase, mock
from datetime import datetime
from werkzeug import MultiDict
from wtforms import StringField
from . import csrf


class TestCSRFForm(TestCase):
    """Tests for :class:`arxiv.forms.csrf.CSRFForm`."""

    @mock.patch(f'{csrf.__name__}.request')
    @mock.patch(f'{csrf.__name__}.get_application_config')
    def test_invalid_token(self, mock_get_config, mock_request):
        """An invalid CSRF token is passed in the form data."""
        mock_request.remote_addr = '10.10.10.10'
        mock_request.auth = None
        mock_request.session = mock.MagicMock(nonce='foononce123')
        mock_get_config.return_value = {'CSRF_SECRET': 'foosecret'}

        class ProtectedForm(csrf.CSRFForm):
            """A CSRF-protected form."""

            something_sensitive = StringField('Something sensitive')

        data = MultiDict({'something_sensitive': 'foo', 'csrf_token': 'nope'})
        form = ProtectedForm(data)
        self.assertFalse(form.validate(), "The form is not valid")

    @mock.patch(f'{csrf.__name__}.request')
    @mock.patch(f'{csrf.__name__}.get_application_config')
    def test_valid_token(self, mock_get_config, mock_request):
        """A valid CSRF token is passed in the form data."""
        ip_address = '10.10.10.10'
        nonce = 'foononce123'
        secret = 'foosecret'
        mock_request.remote_addr = ip_address
        mock_request.auth = None
        mock_request.session = mock.MagicMock(nonce=nonce)
        mock_get_config.return_value = {'CSRF_SECRET': secret}

        class ProtectedForm(csrf.CSRFForm):
            """A CSRF-protected form."""

            something_sensitive = StringField('Something sensitive')

        form = ProtectedForm()

        expires = csrf.SessionCSRF._new_expiry(form.meta.csrf_timeout)
        digest = csrf.SessionCSRF._hash(secret, nonce, ip_address, expires)
        csrf_token = csrf.SessionCSRF._join(digest, expires)

        data = MultiDict({'something_sensitive': 'foo',
                          'csrf_token': csrf_token})
        form = ProtectedForm(data)
        self.assertTrue(form.validate(), "The form is valid")

    @mock.patch(f'{csrf.__name__}.request')
    @mock.patch(f'{csrf.__name__}.get_application_config')
    def test_expired_token(self, mock_get_config, mock_request):
        """A valid but expired CSRF token is passed in the form data."""
        ip_address = '10.10.10.10'
        nonce = 'foononce123'
        secret = 'foosecret'
        expires = datetime.now().isoformat()
        digest = csrf.SessionCSRF._hash(secret, nonce, ip_address, expires)
        csrf_token = csrf.SessionCSRF._join(digest, expires)

        mock_request.remote_addr = ip_address
        mock_request.auth = None
        mock_request.session = mock.MagicMock(nonce=nonce)
        mock_get_config.return_value = {'CSRF_SECRET': secret}

        class ProtectedForm(csrf.CSRFForm):
            """A CSRF-protected form."""

            something_sensitive = StringField('Something sensitive')

        data = MultiDict({'something_sensitive': 'foo',
                          'csrf_token': csrf_token})
        form = ProtectedForm(data)
        self.assertFalse(form.validate(), "The form is not valid")



class TestCSRFFormWithNewSessionRef(TestCase):
    """Test using the ``request.auth`` session ref in accounts v0.4.1."""

    @mock.patch(f'{csrf.__name__}.request')
    @mock.patch(f'{csrf.__name__}.get_application_config')
    def test_invalid_token(self, mock_get_config, mock_request):
        """An invalid CSRF token is passed in the form data."""
        mock_request.remote_addr = '10.10.10.10'
        mock_request.session = None
        mock_request.auth = mock.MagicMock(nonce='foononce123')
        mock_get_config.return_value = {'CSRF_SECRET': 'foosecret'}

        class ProtectedForm(csrf.CSRFForm):
            """A CSRF-protected form."""

            something_sensitive = StringField('Something sensitive')

        data = MultiDict({'something_sensitive': 'foo', 'csrf_token': 'nope'})
        form = ProtectedForm(data)
        self.assertFalse(form.validate(), "The form is not valid")

    @mock.patch(f'{csrf.__name__}.request')
    @mock.patch(f'{csrf.__name__}.get_application_config')
    def test_valid_token(self, mock_get_config, mock_request):
        """A valid CSRF token is passed in the form data."""
        ip_address = '10.10.10.10'
        nonce = 'foononce123'
        secret = 'foosecret'
        mock_request.remote_addr = ip_address
        mock_request.session = None
        mock_request.auth = mock.MagicMock(nonce=nonce)
        mock_get_config.return_value = {'CSRF_SECRET': secret}

        class ProtectedForm(csrf.CSRFForm):
            """A CSRF-protected form."""

            something_sensitive = StringField('Something sensitive')

        form = ProtectedForm()

        expires = csrf.SessionCSRF._new_expiry(form.meta.csrf_timeout)
        digest = csrf.SessionCSRF._hash(secret, nonce, ip_address, expires)
        csrf_token = csrf.SessionCSRF._join(digest, expires)

        data = MultiDict({'something_sensitive': 'foo',
                          'csrf_token': csrf_token})
        form = ProtectedForm(data)
        self.assertTrue(form.validate(), "The form is valid")

    @mock.patch(f'{csrf.__name__}.request')
    @mock.patch(f'{csrf.__name__}.get_application_config')
    def test_expired_token(self, mock_get_config, mock_request):
        """A valid but expired CSRF token is passed in the form data."""
        ip_address = '10.10.10.10'
        nonce = 'foononce123'
        secret = 'foosecret'
        expires = datetime.now().isoformat()
        digest = csrf.SessionCSRF._hash(secret, nonce, ip_address, expires)
        csrf_token = csrf.SessionCSRF._join(digest, expires)

        mock_request.remote_addr = ip_address
        mock_request.session = None
        mock_request.auth = mock.MagicMock(nonce=nonce)
        mock_get_config.return_value = {'CSRF_SECRET': secret}

        class ProtectedForm(csrf.CSRFForm):
            """A CSRF-protected form."""

            something_sensitive = StringField('Something sensitive')

        data = MultiDict({'something_sensitive': 'foo',
                          'csrf_token': csrf_token})
        form = ProtectedForm(data)
        self.assertFalse(form.validate(), "The form is not valid")



r"""
WSGI middleware integration for Flask apps.

A WSGI middleware sits "between" the WSGI server (e.g. uWSGI) and the Flask
application. This allows us to execute code on each request, either before or
after the request is handled by Flask application. For example, a middleware
could be used to parse and validate authorization information before a request
is handled. In practical terms, a middleware is a wrapper around a Flask
application instance.

Writing a middleware
====================

Middlewares may subclass :class:`.base.BaseMiddleware`, which implements some
of the boilerplate needed to make the middleware work with Flask. In that case,
the middleware class need only implement one or both of:

- ``before(environ: dict, start_response: Callable) -> Tuple[dict, Callable]``
- ``after(response: Iterable) -> Iterable``

For example:

.. code-block:: python

   from arxiv.base.middleware import BaseMiddleware

   class FooMiddleware(BaseMiddleware):
       '''Adds the parameter ``foo`` to the request environment.'''

       def before(self, environ: dict, start_response: Callable) \
               -> Tuple[dict, Callable]:
           '''Insert ``foo`` into the environment, and handle the request.'''
           environ['foo'] = 'bar'
           return environ, start_response


In the example above, the ``'foo'`` parameter would be available on the
:prop:`.Flask.request.environ` object within the Flask application.

.. code-block:: python

   from flask import request

   @app.route('/')
   def my_route():
       foo = request.environ['foo']
       return f'The value of foo is {foo}'

For more information, see the `WSGI spec
<https://www.python.org/dev/peps/pep-0333/>`_.

Adding a middleware to a project
================================
This module provides a function called :func:`.wrap` that applies a list of
middlewares to a :class:`.Flask` application. This works by instantiating a
middleware class with a reference to the Flask app, and then replacing the
app's ``wsgi_app`` property with the middleware. In this sense, the middleware
wraps the Flask application.

:func:`.wrap` applies middlewares in reverse order, which means that the
first middleware will be the "outermost" middleware, and will therefore be
called first upon each request.

.. code-block:: python

   from arxiv.base.middleware import wrap
   app = Flask('some_app')
   wrap(app, [FirstMiddleware, SecondMiddleware, ThirdMiddleware])



"""

from typing import Type, Callable, List, Union
import warnings
from flask import Flask

from .base import BaseMiddleware, IWSGIMiddlewareFactory, IWSGIApp
from .. import logging

logger = logging.getLogger(__name__)


def wrap(app: Flask, middlewares: List[IWSGIMiddlewareFactory]) -> Callable:
    """
    Wrap a :class:`.Flask` app in WSGI middlewares.

    Adds/updates ``app.middlewares: Dict[str, IWSGIApp]`` so that middleware
    instances can be accessed later on. Keys are the ``__name__``s of the
    middleware class/factory.

    Parameters
    ----------
    app : :class:`.Flask`
        The Flask app to wrap.
    middlewares : list
        A list of middleware classes. These are applied in reverse, so that the
        first middleware is the "outermost" wrapper around the base ``app``,
        and is therefore called first.

    Returns
    -------
    :class:`.Flask`
        The original Flask ``app``, with middlewares applied.

    """
    if not hasattr(app, 'wsgi_app'):
        raise TypeError('Not a valid Flask app or middleware')

    if not hasattr(app, 'middlewares'):
        app.middlewares = {}

    # Apply the last middleware first, so that the first middleware is called
    # first upon the request.
    wrapped_app: IWSGIApp = app.wsgi_app
    for middleware in middlewares[::-1]:
        try:
            wrapped_app = middleware(wrapped_app, config=app.config)
        except TypeError as e:
            # Maintain backward compatibility with middlewares that don't
            # accept kwargs.
            logger.debug('Encountered TypeError while initializing'
                         ' midleware: %s', e)
            warnings.warn('Middlewares that do not accept kwargs are'
                          ' deprecated. You should update your middleware'
                          ' to accept arbitrary kwargs', DeprecationWarning)
            wrapped_app = middleware(wrapped_app)

        key = getattr(middleware, '__name__', str(middleware))
        app.middlewares[key] = wrapped_app

    app.wsgi_app = wrapped_app  # type: ignore
    return app



"""Provides a convenience class for working with arXiv categories."""

from .definitions import CATEGORIES, ARCHIVES_SUBSUMED, CATEGORY_ALIASES, \
    ARCHIVES, GROUPS


class Category(str):
    """Represents an arXiv category."""

    lookup = CATEGORIES

    @property
    def id(self) -> str:
        """Short name for category ID."""
        return self

    @property
    def name(self) -> str:
        """Get the full category name."""
        if self in self.lookup:
            return str(self.lookup[self]['name'])
        raise ValueError('No such category')

    @property
    def canonical(self) -> 'Category':
        """Get the canonicalized category, if there is one."""
        if self in CATEGORY_ALIASES:
            return Category(CATEGORY_ALIASES[self])
        if self in ARCHIVES_SUBSUMED:
            return Category(ARCHIVES_SUBSUMED[self])
        return self

    @property
    def display(self) -> str:
        """
        Output the display string for a category.

        Examples
        --------
        Earth and Planetary Astrophysics (astro-ph.EP)

        """
        if self in self.lookup:
            catname = self.lookup[self]['name']
            return f'{catname} ({self})'

        parts = self.split('.', 2)
        if len(parts) == 2:    # Has subject.
            (archive, _) = parts
            if archive in ARCHIVES:
                archive_name = ARCHIVES[archive]['name']
                return f'{archive_name} ({archive})'
        return self

    def unalias(self) -> 'Category':
        """Follow any EQUIV or SUBSUMED to get the current category."""
        if self in CATEGORY_ALIASES:
            return Category(CATEGORY_ALIASES[self])
        if self in ARCHIVES_SUBSUMED:
            return Category(ARCHIVES_SUBSUMED[self])
        return self


class Archive(Category):
    """Represents an arXiv archive--the middle level of the taxonomy."""

    lookup = ARCHIVES


class Group(Category):
    """Represents an arXiv group--the highest (most general) taxonomy level."""

    lookup = GROUPS



"""Install arXiv-base as an importable package."""

from setuptools import setup, find_packages


setup(
    name='arxiv-base',
    version='0.16.3',
    packages=[f'arxiv.{package}' for package
              in find_packages('arxiv', exclude=['*test*'])],
    zip_safe=False,
    install_requires=[
        'flask',
        'jsonschema',
        'pytz',
        'uwsgi',
        'boto3',
        'bleach==3.1.0',
        'backports-datetime-fromisoformat==1.0.0',
        'typing-extensions'
    ],
    include_package_data=True
)



"""Tools for encoding/serializing data."""

from typing import Any, Union, List
import json
from datetime import datetime, date
from backports.datetime_fromisoformat import MonkeyPatch
MonkeyPatch.patch_fromisoformat()


class ISO8601JSONEncoder(json.JSONEncoder):
    """Renders date and datetime objects as ISO8601 datetime strings."""

    def default(self, obj: Any) -> Union[str, List[Any]]:
        """Overriden to render date(time)s in isoformat."""
        try:
            if isinstance(obj, (date, datetime)):
                return obj.isoformat()
            iterable = iter(obj)
        except TypeError:
            pass
        else:
            return list(iterable)
        return json.JSONEncoder.default(self, obj)  # type: ignore


class ISO8601JSONDecoder(json.JSONDecoder):
    """Attempts to parse ISO8601 strings as datetime objects."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Pass :func:`object_hook` to the base constructor."""
        kwargs['object_hook'] = kwargs.get('object_hook', self.object_hook)
        super(ISO8601JSONDecoder, self).__init__(*args, **kwargs)

    def _try_isoparse(self, value: Any) -> Any:
        """Attempt to parse a value as an ISO8601 datetime."""
        if type(value) is not str:
            return value
        try:
            # Switched from dateutil.parser because it was too liberal.
            return datetime.fromisoformat(value)  # type: ignore
        except ValueError:
            return value

    def object_hook(self, data: dict, **extra: Any) -> Any:
        """Intercept and coerce ISO8601 strings to datetimes."""
        for key, value in data.items():
            if type(value) is list:
                data[key] = [self._try_isoparse(v) for v in value]
            else:
                data[key] = self._try_isoparse(value)
        return data


def dumps(obj: Any) -> str:
    """Generate JSON from a Python object."""
    return json.dumps(obj, cls=ISO8601JSONEncoder)


def loads(data: str) -> Any:
    """Load a Python object from JSON."""
    return json.loads(data, cls=ISO8601JSONDecoder)
